#!/usr/bin/env python
'''picoscope ps2000 library interface'''

#### SYSTEM IMPORTS ####
import sys
import time
from functools import partial
from ctypes import *  # shouldn't import *, but there are so many items to use...
from array import array

# requires installation of the picoscope software
# (or manually putting the library in the right location)
if sys.platform == 'win32':
    LIBNAME = 'ps2000.dll'
else:
    LIBNAME = '/usr/local/lib/libps2000.so.2.0.7'

PS_MAX_SAMPLES = 8064  # seems to be the max value the scope can do in a block
PS_MAX_VALUE = 32767.0
OVERSAMPLE = 0

# Important to know!! These ranges assume you're using a probe of x1

# channel range values/codes
RANGE_20MV  = 1  # 20 mV
RANGE_50MV  = 2  # 50 mV
RANGE_100MV = 3  # 100 mV
RANGE_200MV = 4  # 200 mV
RANGE_500MV = 5  # 500 mV
RANGE_1V    = 6  # 1 V
RANGE_2V    = 7  # 2 V
RANGE_5V    = 8  # 5 V
RANGE_10V   = 9  # 10 V
RANGE_20V   = 10 # 20 V

# map the range the the scale factor
RANGE_SCALE_MAP = {
        RANGE_20MV  : 0.02,
RANGE_50MV  : 0.05,
RANGE_100MV : 0.1,
RANGE_200MV : 0.2,
RANGE_500MV : 0.5,
RANGE_1V    : 1.0,
RANGE_2V    : 2.0,
RANGE_5V    : 5.0,
RANGE_10V   : 10.0,
RANGE_20V   : 20.0,
}

# probe tip lights
LIGHT_OFF = 0
LIGHT_ON = 1
LIGHT_STATES = [LIGHT_ON, LIGHT_OFF]

# LED states
LED_RED = 1
LED_GREEN = 2
LED_OFF = 3
LED_STATES = [LED_RED, LED_GREEN, LED_OFF]

# channels
CHANNEL_A = 0
CHANNEL_B = 1
CHANNEL_NONE = 5
CHANNELS = [CHANNEL_A, CHANNEL_B, CHANNEL_NONE]

# coupling
COUPLING_DC = 1
COUPLING_AC = 0
COUPLINGS = [COUPLING_DC, COUPLING_AC]


# trigger direction
DIR_FALLING = 1
DIR_RISING = 0
DIRECTIONS = [DIR_RISING, DIR_FALLING]


INFO_NAMES = [
        'DRIVER_VERSION',
        'USB_VERSION',
        'HARDWARE_VERSION',
        'VARIANT_INFO',
        'BATCH_AND_SERIAL',
        'CAL_DATE',
        'ERROR_CODE',
        ]



class PicoError(Exception):
    '''pico scope error'''


# TODO: add checks in methods to ensure lib, handle are valid
class PicoWrapper(object):
    '''picoscope ps2000 interface'''

    def __init__(self):
        self.handle = None

        # load the library
        if sys.platform == 'win32':
            self.lib = windll.LoadLibrary(LIBNAME)
        else:
            self.lib = cdll.LoadLibrary(LIBNAME)
        if not self.lib:
            raise PicoError('could not open library: %s' % LIBNAME)


    ################## low level methods #####################
    def __getattr__(self, name):
        '''this method will call methods starting with ps2000_ via the library'''

        # get a handle to the requested method if name looks like a PS2000 method
        if name.lower().startswith('ps2000_'):
            try:
                func = getattr(self.lib, name)
            except AttributeError:
                raise PicoError('Library "%s" does not support method "%s"' % (LIBNAME, func_name))
            else:
                # return a partial function of the library method with handle passed in
                return partial(func, self.handle)

        # not a ps2000 request, defer
        else:
            raise AttributeError


    def ps2000_close_unit(self):
        '''low-level close the interface to the unit'''
        res = self.lib.ps2000_close_unit(self.handle)
        self.handle = None
        return res


    def ps2000_open_unit(self):
        '''low-level open interface to unit'''
        self.handle = self.lib.ps2000_open_unit()
        return self.handle


    ################## higher level methods #####################
    def close_unit(self):
        '''close the unit'''
        res = self.ps2000_close_unit()
        if res == 0:
            raise PicoError('flash_led: invalid handle')


    def flash_led(self):
        '''flash led on front of unit'''
        res = self.ps2000_flash_led()
        if res == 0:
            raise PicoError('flash_led: invalid handle')


#ps2000_get_streaming_last_values
#ps2000_get_streaming_values
#ps2000_get_streaming_values_no_aggregation

    def get_timebase(self, timebase, num_samples, oversample):
        '''return the (time_interval, time_units, max_samples) for the given parameters'''

        time_interval = c_long()
        time_units = c_short()
        max_samples = c_long()
        res = self.ps2000_get_timebase(timebase, num_samples, byref(time_interval),
                                       byref(time_units), oversample, byref(max_samples))
        if res == 0:
            raise PicoError('get_timebase: parameter out of range')
        else:
            return time_interval.value, time_units.value, max_samples.value


    def get_times_and_values(self, time_units, num_points):
        '''return requested amount of data (times, chA, chB, overflowA, overflowB)'''
        times =     (c_long * num_points)()
        ch_a_data = (c_short * num_points)()
        ch_b_data = (c_short * num_points)()
        overflow =  c_short()
        res = self.ps2000_get_times_and_values(byref(times), byref(ch_a_data), byref(ch_b_data),
                                               None, None, byref(overflow), time_units, num_points)
 #       if res == 0:
 #           raise PicoError('get_times_and_values: failed!')
 #       else:
        return array('l', times), array('h', ch_a_data), array('h', ch_b_data)



    def get_unit_info(self):
        '''return information about the device'''
        info = ''
        for line, name in enumerate(INFO_NAMES):
            buf = create_string_buffer('\000'*1024)
            res = self.ps2000_get_unit_info(buf, 1024, line)
            if res == 0:
                raise PicoError('get_unit_info: failed!')
            else:
                info += '%s:  %s\n' % (name,  buf.value)

        return info




    def get_values(self, num_points):
        '''return requested amount of data (chA, chB, overflowA, overflowB)'''
        ch_a_data = (c_short * num_points)()
        ch_b_data = (c_short * num_points)()
        overflow =  c_short()
        res = self.ps2000_get_values(byref(ch_a_data), byref(ch_b_data),
                                     None, None, byref(overflow), num_points)
        if res == 0:
            raise PicoError('get_values: failed!')
        else:
            return array('h', ch_a_data), array('h', ch_b_data)


#ps2000_last_button_press


    def open_unit(self):
        '''open interface to unit'''
        res = self.ps2000_open_unit()
        if res < 0:
            raise PicoError('open: failed to open scope')
        elif res == 0:
            raise PicoError('open: could not find picoscope unit')


#ps2000_open_unit_async
#ps2000_open_unit_progress
#ps2000_overview_buffer_status


    def ready(self):
        '''indicate if previous block acquire is complete - returns bool'''
        res = self.ps2000_ready()
        if res < 1:
            raise PicoError('ready: oscilloscope not attached')
        else:
            return bool(res)


    def run_block(self, num_samples, timebase, oversample):
        '''start acquisition of block of data; return expected duration of acquisition in mS'''
        time_indisposed_ms = c_long()

        res = self.ps2000_run_block(num_samples, timebase, oversample, byref(time_indisposed_ms))
        if res == 0:
            raise PicoError('run_block: parameter out of range')
        else:
            return time_indisposed_ms.value


#ps2000_run_streaming
#ps2000_run_streaming_ns
#ps2000SetAdvTriggerChannelConditions
#ps2000SetAdvTriggerChannelDirections
#ps2000SetAdvTriggerChannelProperties
#ps2000SetAdvTriggerDelay


    def set_channel(self, channel, enabled, coupling, range):
        '''select channel and modes'''
        res = self.ps2000_set_channel(channel, enabled, coupling, range)

        if res == 0:
            raise PicoError('set_channel: failed')

#ps2000_set_ets


    def set_led(self, state):
        '''set colour/state of led on scope'''
        if state not in LED_STATES:
            raise PicoError('set_led: unknown led state: %s' % state)

        res = self.ps2000_set_led(state)

        if res == 0:
            raise PicoError('set_led: failed')


    def set_light(self, state):
        '''enable/disable light on probe tip'''
        if state not in LIGHT_STATES:
            raise PicoError('set_light: unknown light state: %s' % state)

        res = self.ps2000_set_light(state)

        if res == 0:
            raise PicoError('set_light: failed')


#ps2000SetPulseWidthQualifier
#ps2000SetSigGenArbitrary
#ps2000SetSigGenBuiltIn


    def set_trigger(self, source, threshold, direction, delay, auto_trigger_ms):
        ''' Set the trigger mode. Note, you have to be careful with the latest drivers. It seems as though
            the picoscope will not trigger if the first edge seen is not the direction you tell it to
            trigger on (for example if triggering is set for rising edge and a pulse with falling edge
            followed by rising edge is seen, it will not trigger!).
            source: Where to look for a trigger; CHANNEL_A, CHANNEL_B, CHANNEL_NONE.
            threshold: The threshold for the trigger event (scaled in 16-bit ADC counts)
                       at the currently selected range.
            direction: DIR_RISING, DIR_FALLING
            delay: The delay, as a percentage of the requested number of data points, between the
            trigger event and the start of the block. It should be in the range -100% to +100%.
            Thus, 0% means that the trigger event is at the first data value in the block, and -50%
            means that it is in the middle of the block.
            auto_trigger_ms: the delay in milliseconds after which the oscilloscope will collect
            samples if no trigger event occurs. If this is set to zero the oscilloscope will wait
            for a trigger indefinitely.
        '''

        res = self.ps2000_set_trigger(source, threshold, direction, delay, auto_trigger_ms)

        if res == 0:
            raise PicoError('set_trigger: failed')


    def stop(self):
        '''stop scope acquisition'''
        res = self.ps2000_stop()

        if res == 0:
            raise PicoError('stop: failed')




class PS2000(PicoWrapper):
    '''PS2000 class utilizing the PicoWrapper with higher level functionality'''

    def __init__(self):
        PicoWrapper.__init__(self)

        self.ranges = {CHANNEL_A: None, CHANNEL_B: None}
        self.selected_timebase = None

        # store the available timebases
        self.timebases = {}


    def open_unit(self):
        '''overridden method to store timebases'''
        super(PS2000, self).open_unit()

        # store the timebases
        for i in range(100): # ??
            try:
                interval, units, samples = self.get_timebase(i, PS_MAX_SAMPLES, 0)
                assert samples == PS_MAX_SAMPLES  # we only support using max samples
                self.timebases[i] = (interval, units, interval * samples) # last one is time per frame in nS
            except PicoError:
                pass


    def set_minimum_time_per_frame(self, time_in_ns):
        '''set the timebase by specifying the time in nS per frame, return the actual time per frame'''
        self.selected_timebase = None

        for timebase, (interval, units, time_per_frame_ns) in self.timebases.items():
            if time_per_frame_ns >= time_in_ns:
                self.selected_timebase = timebase
                return time_per_frame_ns
        else:
            raise PicoError('Could not find timebase sufficient for %d nS' % time_in_ns)




    @property
    def time_per_frame_ns(self):
        '''return the amount of time per frame at the current timebase'''
        return self.timebases[self.selected_timebase][2]


    def get_data(self, scaleA=1.0, scaleB=1.0):
        '''blocking call to wait for data; return the data as values from 0 to 1, then scaled to the voltage range'''
        while not self.ready():
            pass  # TODO: sleep
        a, b = self.get_values(PS_MAX_SAMPLES)
        rangeA = RANGE_SCALE_MAP[self.ranges[CHANNEL_A]]
        rangeB = RANGE_SCALE_MAP[self.ranges[CHANNEL_B]]
        return [scaleA * rangeA * v/PS_MAX_VALUE for v in a], [scaleB * rangeB * v/PS_MAX_VALUE for v in b]


    def get_timed_data(self, scaleA=1.0, scaleB=1.0):
        '''blocking call to wait for data'''
        while not self.ready():
            pass  # TODO: sleep
        units = self.timebases[self.selected_timebase][1]
        t, a, b = self.get_times_and_values(units, PS_MAX_SAMPLES)
        rangeA = RANGE_SCALE_MAP[self.ranges[CHANNEL_A]]
        rangeB = RANGE_SCALE_MAP[self.ranges[CHANNEL_B]]
        return t, [scaleA * rangeA * v/PS_MAX_VALUE for v in a], [scaleB * rangeB * v/PS_MAX_VALUE for v in b]


    def acquire_block(self):
        '''acquisition of num_samples at current sample_interval'''
        self.run_block(PS_MAX_SAMPLES, self.selected_timebase, OVERSAMPLE)
        return self.get_data()


    def start_block(self):
        '''start run_block with current data setup'''
        return self.run_block(PS_MAX_SAMPLES, self.selected_timebase, OVERSAMPLE)


    def set_channel(self, channel, enabled, coupling, _range):
        '''overriden set_channel: used to store the range values for later use'''
        assert channel in self.ranges, 'channel %r, not found in range list: %r' % (channel, self.ranges.keys())

        super(PS2000, self).set_channel(channel, enabled, coupling, _range)
        self.ranges[channel] = _range # done after in case the call fails



    def quick_setup(self, options={}):
        '''convenience routine to capture scope data - caller must call get_timed_data()
           the return value is the actual time per frame - caller must use this for accuracy!!'''
        opts = dict(
                trigger_source=CHANNEL_NONE,
                trigger_thresh=0,
                trigger_dir=DIR_FALLING,
                trigger_delay=0,
                trigger_auto_ms=0,
                chl_A_enabled=True,
                chl_A_range=RANGE_20V,
                chl_A_coupling=COUPLING_DC,
                chl_B_enabled=True,
                chl_B_range=RANGE_20V,
                chl_B_coupling=COUPLING_DC,
                time_per_frame_ns=50e3,
                )

        # update the defauls with what we're given
        opts.update(options)


        self.set_channel(CHANNEL_A,
                               opts['chl_A_enabled'],
                               opts['chl_A_coupling'],
                               opts['chl_A_range'])
        self.set_channel(CHANNEL_B,
                               opts['chl_B_enabled'],
                               opts['chl_B_coupling'],
                               opts['chl_B_range'])
        self.set_trigger(opts['trigger_source'],
                               opts['trigger_thresh'],
                               opts['trigger_dir'],
                               opts['trigger_delay'],
                               opts['trigger_auto_ms'])
        time_per_frame_ns = self.set_minimum_time_per_frame(opts['time_per_frame_ns'])
        self.start_block()
        return time_per_frame_ns



################################################################################
if __name__ == '__main__':

    # example code
    NUM_SAMPLES = 100

    #print 'creating'
    scope = PS2000()
    print 'open'
    scope.open_unit()
    print 'info'
    print scope.get_unit_info()
    #print 'set channel'
    scope.set_channel(CHANNEL_A, True, COUPLING_DC, RANGE_5V)
    scope.set_channel(CHANNEL_B, True, COUPLING_DC, RANGE_5V)
    #print 'set trigger'
    scope.set_trigger(CHANNEL_NONE, 0, DIR_FALLING, 0, 0)

    #print scope.set_trigger(CHANNEL_B, 0x2000, DIR_FALLING, -5, 0) # TODO: fix threshold
    #print 'set timebase'
    scope.set_timebase(80000000)
    print 'start_block'
    scope.start_block()
    time.sleep(0.1)
    print 'get_data'
    data = scope.get_timed_data()
    print 'done'

#    import aplotter
#    aplotter.plot(data[2])


    # start the interactive prompt
    try:
        from IPython.frontend.terminal.embed import InteractiveShellEmbed
        ipshell = InteractiveShellEmbed()

    except:  # versions older than 0.12
        from IPython.Shell import IPShellEmbed
        ipshell = IPShellEmbed([])  # pass empty argument list

    ipshell()

    sys.exit(1)
